#include "Cv11_main.h"

//---------------------
//----    SCENE    ----
//---------------------

/// Reloads all shaders
void reload_shaders()
{
	depth_prepass_program.Init();
	depth_prepass_program.AddVertexShader("Shaders/nolit_vertex.glsl");
	depth_prepass_program.AddFragmentShader("Shaders/nothing_fragment.glsl");
	depth_prepass_program.Link();

	skybox_program.Init();
	skybox_program.AddVertexShader("Shaders/skybox_vertex.glsl");
	skybox_program.AddFragmentShader("Shaders/skybox_fragment.glsl");
	skybox_program.Link();

	object_program.Init();
	object_program.AddVertexShader("Shaders/object_vertex.glsl");
	object_program.AddFragmentShader("Shaders/object_fragment.glsl");
	object_program.Link();

	cout << "Shaders are reloaded" << endl;
}

/// Initializes all objects of our scene
void init_scene()
{
	//----------------------------------------------
	//--  Display some information about the OpenGL

	PrintOpenGLInfo();

	//----------------------------------------------
	//--  Load shaders

	reload_shaders();
	
	//----------------------------------------------
	//--  Prepare all cameras
	
	// Create the basic camera through which we look into the scene
	the_camera.SetCenter(glm::vec3(0.0f));
	the_camera.SetCamera(-0.5f, -0.5f, 15.0f);
	CameraData_ubo.Init();

	//----------------------------------------------
	//--  Prepare geometries
	
	geom_cube = CreateCube();
	geom_sphere = CreateSphere();
	geom_torus = CreateTorus();
	geom_cylinder = CreateCylinder();
	geom_capsule = CreateCapsule();
	geom_teapot = CreateTeapot();
	geom_fullscreen_quad = CreateFullscreenQuad();

	//----------------------------------------------
	//--  Load textures and set their parameters

	// Load mask texture
	mask_tex = CreateAndLoadTexture2D(L"../../textures/Mask.png");
	glBindTexture(GL_TEXTURE_2D, mask_tex);
	SetTexture2DParameters(GL_REPEAT, GL_REPEAT, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	// Load meadow skybox
	meadow_skybox_tex = CreateAndLoadTextureCube(
		L"../../textures/meadow_skybox/meadow_skybox_px.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_nx.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_py.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_ny.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_pz.jpg",
		L"../../textures/meadow_skybox/meadow_skybox_nz.jpg");
	glBindTexture(GL_TEXTURE_CUBE_MAP, meadow_skybox_tex);
	SetTextureCubeParameters(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAX_ANISOTROPY_EXT, 16.0f);
	glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
	
	// Load miramar skybox
	miramar_skybox_tex = CreateAndLoadTextureCube(
		L"../../textures/miramar_skybox/miramar_skybox_px.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_nx.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_py.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_ny.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_pz.jpg",
		L"../../textures/miramar_skybox/miramar_skybox_nz.jpg");
	glBindTexture(GL_TEXTURE_CUBE_MAP, miramar_skybox_tex);
	SetTextureCubeParameters(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAX_ANISOTROPY_EXT, 16.0f);
	glGenerateMipmap(GL_TEXTURE_CUBE_MAP);
	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

	// Choose initial skybox texture
	current_skybox_tex = meadow_skybox_tex;

	//----------------------------------------------
	//--  Positions of objects in the scene

	float x_start = -14.0f;
	float z_start = -14.0f;
	float x_spacing = 4.0f;
	float z_spacing = 4.0f;
	int x_grid_size = 8;
	int z_grid_size = 8;

	for (int i = 0; i < 64; i++)
	{
		CentralObjects_ubo[i].Init();
		CentralObjects_ubo[i].SetMatrix(glm::translate(glm::mat4(1.0f),
			glm::vec3(x_start + float(i%x_grid_size) * x_spacing, 0.0f, z_start + float(i/z_grid_size) * z_spacing)));
		CentralObjects_ubo[i].UpdateOpenGLData();

		switch (i%8)
		{
		case 0:		material_roughnesses[i] = 0.00f;		break;
		case 1:		material_roughnesses[i] = 0.05f;		break;
		case 2:		material_roughnesses[i] = 0.10f;		break;
		case 3:		material_roughnesses[i] = 0.15f;		break;
		case 4:		material_roughnesses[i] = 0.20f;		break;
		case 5:		material_roughnesses[i] = 0.30f;		break;
		case 6:		material_roughnesses[i] = 0.40f;		break;
		case 7:		material_roughnesses[i] = 0.50f;		break;
		}
		switch (i/8)
		{
		case 0:		material_fresnels[i] = glm::vec3(0.04f, 0.04f, 0.04f);		material_diffuses[i] = glm::vec3(1.0f, 0.0f, 0.0f);		break;		// Plastic Red
		case 1:		material_fresnels[i] = glm::vec3(0.04f, 0.04f, 0.04f);		material_diffuses[i] = glm::vec3(0.0f, 1.0f, 0.0f);		break;		// Plastic Green
		case 2:		material_fresnels[i] = glm::vec3(0.04f, 0.04f, 0.04f);		material_diffuses[i] = glm::vec3(0.0f, 0.0f, 1.0f);		break;		// Plastic Blue
		case 3:		material_fresnels[i] = glm::vec3(0.56f, 0.57f, 0.58f);		material_diffuses[i] = glm::vec3(0.0f, 0.0f, 0.0f);		break;		// Iron
		case 4:		material_fresnels[i] = glm::vec3(0.95f, 0.64f, 0.54f);		material_diffuses[i] = glm::vec3(0.0f, 0.0f, 0.0f);		break;		// Copper
		case 5:		material_fresnels[i] = glm::vec3(1.00f, 0.71f, 0.29f);		material_diffuses[i] = glm::vec3(0.0f, 0.0f, 0.0f);		break;		// Gold
		case 6:		material_fresnels[i] = glm::vec3(0.91f, 0.92f, 0.92f);		material_diffuses[i] = glm::vec3(0.0f, 0.0f, 0.0f);		break;		// Aluminium
		case 7:		material_fresnels[i] = glm::vec3(0.95f, 0.93f, 0.88f);		material_diffuses[i] = glm::vec3(0.0f, 0.0f, 0.0f);		break;		// Silver
		}
	}

	//----------------------------------------------
	//--  Compute the random sampling directions

	srand(12345);
	std::vector<glm::vec4> RandomFloats(1024);
	for (size_t i = 0; i < RandomFloats.size(); i++)
	{
		RandomFloats[i] = glm::vec4(
			float(rand()) / float(RAND_MAX),
			float(rand()) / float(RAND_MAX),
			float(rand()) / float(RAND_MAX),
			float(rand()) / float(RAND_MAX));
	}
	// Create a UBO for these random positions
	glGenBuffers(1, &RandomFloats_UBO);
	glBindBuffer(GL_UNIFORM_BUFFER, RandomFloats_UBO);
	glBufferData(GL_UNIFORM_BUFFER, sizeof(float) * 4 * RandomFloats.size(), RandomFloats.data(), GL_STATIC_DRAW);
	glBindBuffer(GL_UNIFORM_BUFFER, 0);

	//----------------------------------------------
	//--  Miscelaneous

	// Create the query object
	glGenQueries(1, &RenderTimeQuery);
}

/// Updates the scene: performs animations, updates the data of the buffers, etc.
void update_scene(int app_time_diff_ms)
{
	//----------------------------------------------
	//--  Choose the correct skybox
	
	if (skybox_texture == SKYBOX_MEADOW)
	{
		current_skybox_tex = meadow_skybox_tex;
	}
	else if (skybox_texture == SKYBOX_MIRAMAR)
	{
		current_skybox_tex = miramar_skybox_tex;
	}

	//----------------------------------------------
	//--  Data of the main camera
	
	CameraData_ubo.SetProjection(
		glm::perspective(glm::radians(45.0f), float(win_width) / float(win_height), 0.1f, 1000.0f));
	CameraData_ubo.SetCamera(the_camera);
	CameraData_ubo.UpdateOpenGLData();
}

/// Renders the whole frame
void render_scene()
{
	// Start measuring the elapsed time
	glBeginQuery(GL_TIME_ELAPSED, RenderTimeQuery);

	// Render into the main window
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glViewport(0, 0, win_width, win_height);

	// Clear the framebuffer, clear the color and depth
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);

	// Set the data of the camera and the lights
	CameraData_ubo.BindBuffer(DEFAULT_CAMERA_BINDING);
	glBindBufferBase(GL_UNIFORM_BUFFER, 4, RandomFloats_UBO);
	glActiveTexture(GL_TEXTURE0);		glBindTexture(GL_TEXTURE_CUBE_MAP, current_skybox_tex);
	glActiveTexture(GL_TEXTURE1);		glBindTexture(GL_TEXTURE_2D, mask_tex);

	// Render the skybox
	if (skybox_program.IsValid())
	{
		skybox_program.Use();

		glActiveTexture(GL_TEXTURE0);		glBindTexture(GL_TEXTURE_CUBE_MAP, current_skybox_tex);

		// Render the skybox only into the color buffer, do not use depth test, do not update depth buffer
		glDepthMask(GL_FALSE);
		glDisable(GL_DEPTH_TEST);

		// We render the skybox using a fullscreen quad, and we compute the view direction in vertex shader
		geom_fullscreen_quad.BindVAO();
		geom_fullscreen_quad.Draw();
	
		// Restore the depth test and the depth buffer writes
		glDepthMask(GL_TRUE);
		glEnable(GL_DEPTH_TEST);
	}

	// Choose proper object
	Geometry *geom = nullptr;
	switch (central_object)
	{
	case CENTRAL_OBJECT_CUBE:		geom = &geom_cube;		break;
	case CENTRAL_OBJECT_SPHERE:		geom = &geom_sphere;	break;
	case CENTRAL_OBJECT_TORUS:		geom = &geom_torus;		break;
	case CENTRAL_OBJECT_CYLINDER:	geom = &geom_cylinder;	break;
	case CENTRAL_OBJECT_CAPSULE:	geom = &geom_capsule;	break;
	case CENTRAL_OBJECT_TEAPOT:		geom = &geom_teapot;	break;
	}

	// Depth prepass - render only into depth buffer
	glDepthFunc(GL_LEQUAL);
	if (depth_prepass_program.IsValid())
	{
		depth_prepass_program.Use();
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

		geom->BindVAO();
		for (int i = 0; i < 64; i++)
		{
			CentralObjects_ubo[i].BindBuffer(DEFAULT_OBJECT_BINDING);
			geom->Draw();
		}

		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	}

	// Render all objects in the scene
	glDepthFunc(GL_EQUAL);
	if (object_program.IsValid())
	{
		object_program.Use();
		object_program.Uniform1i("diffuse_sample_count", diffuse_sample_count);
		object_program.Uniform1i("specular_sample_count", specular_sample_count);
		object_program.Uniform1i("mask_roughness", mask_roughness ? 1 : 0);
		object_program.Uniform1i("use_shiny_layer", use_shiny_layer ? 1 : 0);
		object_program.Uniform1i("use_shiny_layer_mask", use_shiny_layer_mask ? 1 : 0);

		geom->BindVAO();
		for (int i = 0; i < 64; i++)
		{
			object_program.Uniform3fv("material_fresnel", 1, glm::value_ptr(material_fresnels[i]));
			object_program.Uniform3fv("material_diffuse", 1, glm::value_ptr(material_diffuses[i]));
			object_program.Uniform1f("material_roughness", material_roughnesses[i]);
			CentralObjects_ubo[i].BindBuffer(DEFAULT_OBJECT_BINDING);
			geom->Draw();
		}
	}

	// Reset the VAO and the program
	glBindVertexArray(0);
	glUseProgram(0);
	glBindFramebuffer(GL_FRAMEBUFFER, 0);

	// Stop measuring the elapsed time
	glEndQuery(GL_TIME_ELAPSED);

	// Evaluate the query
	glFinish();					// Wait for OpenGL, don't forget OpenGL is asynchronous
	GLuint64 render_time;
	glGetQueryObjectui64v(RenderTimeQuery, GL_QUERY_RESULT, &render_time);
	render_time_ms = float(render_time) * 1e-6f;
}

//-------------------
//----    GUI    ----
//-------------------

/// Callback when user clicks on Reload button
void TW_CALL reload(void *)
{
	reload_shaders();
}

void init_gui()
{
	central_object = CENTRAL_OBJECT_CUBE;
	skybox_texture = SKYBOX_MEADOW;

	diffuse_sample_count = 64;
	specular_sample_count = 128;
	mask_roughness = false;
	use_shiny_layer = false;
	use_shiny_layer_mask = false;

	// Initialize AntTweakBar
	TwInit(TW_OPENGL_CORE, nullptr);
	TwWindowSize(win_width, win_height);
	TwGLUTModifiersFunc(glutGetModifiers);

	// Initialize GUI
	the_gui = TwNewBar("Parameters");
	TwDefine("Parameters size='270 400'");
	TwDefine("Parameters valueswidth=100");
	TwAddButton(the_gui, "Reload", reload, nullptr, nullptr);

	// Render time (ms)
	TwAddVarRO(the_gui, "Render time (ms)", TW_TYPE_FLOAT, &render_time_ms, nullptr);

	// Skybox
	TwEnumVal skyboxes[2] = {
		{SKYBOX_MEADOW, "Meadow"},
		{SKYBOX_MIRAMAR, "Miramar"},
	};
	TwType skyboxType = TwDefineEnum("Skyboxes", skyboxes, 2);
	TwAddVarRW(the_gui, "Skybox", skyboxType, &skybox_texture, nullptr);

	// Central object
	TwEnumVal objects[6] = {
		{CENTRAL_OBJECT_CUBE, "Cube"},
		{CENTRAL_OBJECT_SPHERE, "Sphere"},
		{CENTRAL_OBJECT_TORUS, "Torus"},
		{CENTRAL_OBJECT_CYLINDER, "Cylinder"},
		{CENTRAL_OBJECT_CAPSULE, "Capsule"},
		{CENTRAL_OBJECT_TEAPOT, "Teapot"},
	};
	TwType objectType = TwDefineEnum("Objects", objects, 6);
	TwAddVarRW(the_gui, "Central object", objectType, &central_object, nullptr);

	TwAddSeparator(the_gui, nullptr, nullptr);

	// Number of samples for diffuse sampling
	TwEnumVal sampleCounts[8] = {
		{32, "32"},
		{64, "64"},
		{128, "128"},
		{256, "256"},
		{384, "384"},
		{512, "512"},
		{768, "768"},
		{1024, "1024"},
	};
	TwType sampleCountsType = TwDefineEnum("SampleCounts", sampleCounts, 8);
	TwAddVarRW(the_gui, "Diffuse sample count", sampleCountsType, &diffuse_sample_count, nullptr);
	TwAddVarRW(the_gui, "Specular sample count", sampleCountsType, &specular_sample_count, nullptr);

	TwAddVarRW(the_gui, "Use MaskTex for roughness", TW_TYPE_BOOLCPP, &mask_roughness, nullptr);
	TwAddVarRW(the_gui, "Use shiny layer", TW_TYPE_BOOLCPP, &use_shiny_layer, nullptr);
	TwAddVarRW(the_gui, "Use MaskTex for shiny layer", TW_TYPE_BOOLCPP, &use_shiny_layer_mask, nullptr);
}

//---------------------------
//----    APPLICATION    ----
//---------------------------

/// GLUT callback - when the user presses a key. See glutKeyboardFunc for more info
void on_keyboard_func(unsigned char key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventKeyboardGLUT(key, x, y))
		return;		// Already handled by AntTweakBar

	switch (key)
	{
	case 27:		// Escape key
		exit(0);
		break;
	}
}

/// GLUT callback - when the user presses a special key (like F1 key). See glutSpecialFunc for more info
void on_special_func(int key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventSpecialGLUT(key, x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the user presses or releases a mouse button. See glutMouseFunc for more info
void on_mouse_func(int button, int state, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseButtonGLUT(button, state, x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMouseFunc(button, state, x, y);
}

/// GLUT callback - when the user moves the mouse when holding a botton. See glutMotionFunc for more info
void on_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMotionFunc(x, y);
}

/// GLUT callback - when the user moves the mouse when not holding a botton. See glutPassiveMotionFunc for more info
void on_passive_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the window is resized. See glutReshapeFunc for more info
void on_reshape(int width, int height)
{
	win_width = width;
	win_height = height;
	
	glViewport(0, 0, win_width, win_height);
	
	// Inform AntTweakBar
	TwWindowSize(win_width, win_height);
}

/// GLUT callback - when our animation timer requests a new frame. See glutTimerFunc for more info
void on_timer(int)
{
	// Register the timer to be called again in 20 ms, and rerender the scene
	glutTimerFunc(20, on_timer, 0);
	glutPostRedisplay();
}

/// GLUT callback - when the window needs to be redrawn. See glutDisplayFunc for more info
void on_display()
{
	//--  Update all the data

	// Update the application time
	int current_glut_time = glutGet(GLUT_ELAPSED_TIME);
	int app_time_diff_ms = current_glut_time - last_glut_time;
	app_time_ms += app_time_diff_ms;
	last_glut_time = current_glut_time;

	// Update the scene
	update_scene(app_time_diff_ms);

	//--  Render the scene

	// Render the scene
	render_scene();
	
	// Render the GUI
	TwDraw();

	// Swaps the front and back buffer (double-buffering)
	glutSwapBuffers();
}

/// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
	switch (type)			// Report only some errors
	{
	case GL_DEBUG_TYPE_ERROR:
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		cout << message << endl;		// Put the breakpoint here
		return;
	default:
		return;
	}
}

/// C main function :-)
int main(int argc, char ** argv)
{
	// Initialize GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);		// Return from the main loop

	// Set OpenGL Context parameters
	glutInitContextVersion(4, 3);				// Use OpenGL 4.3
	glutInitContextProfile(GLUT_CORE_PROFILE);	// Use OpenGL core profile
	glutInitContextFlags(GLUT_DEBUG);			// Use OpenGL debug context

	// Initialize window
	glutInitWindowSize(win_width, win_height);
	glutCreateWindow("PV227 - Cv11: Image Based Lighting");

	// Initialize GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	// Initialize DevIL library
	InitDevIL();

	// Set the OpenGL debug callback
	SetDebugCallback(simple_debug_callback);

	// Register GLUT callbacks
	glutDisplayFunc(on_display);
	glutReshapeFunc(on_reshape);
	glutTimerFunc(20, on_timer, 0);

	glutKeyboardFunc(on_keyboard_func);
	glutSpecialFunc(on_special_func);
	glutMouseFunc(on_mouse_func);
	glutMotionFunc(on_motion_func);
	glutPassiveMotionFunc(on_passive_motion_func);

	// Initialize OpenGL stuff
	init_gui();
	init_scene();

	// Everything is loaded and prepared, reset the application time
	last_glut_time = glutGet(GLUT_ELAPSED_TIME);
	app_time_ms = 0;

	// Run the main loop
	glutMainLoop();

	// Unload AntTweakBar
	TwTerminate();

	return 0;
}